using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Web.UI;
using System.Web.UI.WebControls;
using Microsoft.SharePoint;
using Microsoft.SharePoint.WebControls;
using NUnit.Framework;
using PS.WebParts.Rollup.Core;
using PS.WebParts.Rollup.Query;

namespace PS.WebParts.Rollup.Test
{
    [TestFixture]
    public class TestAggregateInterpreter
    {
        [Test]
        public void TestWalkSingleJoin()
        {
            DataSet ds = new DataSet();
            DataTable t1;
            DataTable t2;

            SetupTables(ds, out t1, out t2);
            
            Dictionary<string, DataTable> tables = new Dictionary<string, DataTable>(2);
            tables.Add("T1", t1);
            tables.Add("T2", t2);

            Join operation = new Join("T1");
            operation.BaseColumn = "ID";
            operation.JoineeId = "T2";
            operation.JoineeColumn = "ID";
            operation.JoinType = Join.Type.Inner;
            
            AggregateInterpreter interpreter = new AggregateInterpreter(operation, tables);
            DataTable res = interpreter.GetResults();
            
            Assert.AreSame(t1, res);
            Assert.AreEqual(3, res.Rows.Count);
            Assert.AreEqual(6, res.Columns.Count);
        }

        [Test]
        public void TestDatasourceJoin()
        {

            SPDataSource sdp = new SPDataSource();
            sdp.SelectCommand = @"<ViewFields>
<FieldRef Name='Title' />
<FieldRef Name='ShipID' />
</ViewFields>";
            IEnumerable enumerable = sdp.GetView().Select(new DataSourceSelectArguments());
            DataTable dt = enumerable as DataTable;
            Assert.IsNotNull(dt);

            SqlRollupDataSource sqlDs = new SqlRollupDataSource();
            sqlDs.TopLevelSite = "http://localhost/rollup";
            sqlDs.Lists = "TypedList";
            sqlDs.Scope = RollUpQueryWebScope.Web;
            sqlDs.Fields = @"<FieldRef Name='ShipID' /> <FieldRef Name='RecordDate' />";
            sqlDs.ID = "sqlDs";
            
            SPDataSourceWrap spWrap = new SPDataSourceWrap();
            spWrap.ID = "spWrap";
            spWrap.Query =
                @"<ViewFields>
<FieldRef Name='Title' />
<FieldRef Name='ShipID' />
</ViewFields>";

            DataSourceCollection dsc = new DataSourceCollection();
            dsc.Add(sqlDs);
            dsc.Add(spWrap);

            Join operation = new Join("sqlDs");
            operation.BaseColumn = "ShipID";
            operation.JoineeId = "spWrap";
            operation.JoineeColumn = "ShipID";
            operation.JoinType = Join.Type.Inner;

            AggregateInterpreter ai = new AggregateInterpreter(operation, dsc);
            DataTable res = ai.GetResults();
        }

        [Test]
        public void TestSingleUnion()
        {
            DataTable t1;
            DataTable t2;

            SetupTablesForUnion(out t1, out t2);

            Dictionary<string, DataTable> tables = new Dictionary<string, DataTable>(2);
            tables.Add("T1", t1);
            tables.Add("T2", t2);

            Union operation = new Union();
            operation.Children.Add(new DataTableRef("T1") );
            operation.Children.Add(new DataTableRef("T2"));

            AggregateInterpreter interpreter = new AggregateInterpreter(operation, tables);
            DataTable results = interpreter.GetResults();
            
            Assert.AreEqual(4, results.Rows.Count);
            Assert.AreEqual(4, results.Columns.Count);
        }

        private static List<DataTable> SetupTablesForUnion(out DataTable t1, out DataTable t2)
        {
            t1 = new DataTable("T1");
            t2 = new DataTable("T2");

            t1.Columns.Add("ID");
            t1.Columns.Add("A");
            t1.Columns.Add("B");


            t2.Columns.Add("ID");
            t2.Columns.Add("A");
            t2.Columns.Add("C");

            DataRowCollection baseRows = t1.Rows;
            baseRows.Add("1", "a1", "b1");
            baseRows.Add("2", "a2", "b2");

            t2.Rows.Add("1", "a2_1", "c1");
            t2.Rows.Add("2", "a2_3", "c3");

            List<DataTable> tables = new List<DataTable>(2);
            tables.Add(t1);
            tables.Add(t2);
            return tables;
        }

        
        private static DataRowCollection SetupTables(DataSet ds, out DataTable t1, out DataTable t2)
        {
            t1 = new DataTable("T1");
            t2 = new DataTable("T2");

            ds.Tables.Add(t1);
            ds.Tables.Add(t2);

            t1.Columns.Add("ID");
            t1.Columns.Add("A");
            t1.Columns.Add("B");


            t2.Columns.Add("ID");
            t2.Columns.Add("C");
            t2.Columns.Add("D");

            DataRowCollection baseRows = t1.Rows;
            baseRows.Add("1", "a1", "b1");
            baseRows.Add("2", "a2", "b2");
            baseRows.Add("3", "a3", "b3");
            baseRows.Add("4", "a4", "b4");

            t2.Rows.Add("1", "c1", "d1");
            t2.Rows.Add("1", "c2", "d2");
            t2.Rows.Add("2", "c3", "d3");
            return baseRows;
        }
    }
}
